import os
import unittest

from . import consent


class TestFlypaperConsent(unittest.TestCase):
    def setUp(self):
        os.environ["FLYPAPER_TEST_OPT_IN_AFFIRMATIVE"] = "1"
        os.environ["FLYPAPER_TEST_OPT_OUT_AFFIRMATIVE"] = "1"
        os.environ["FLYPAPER_TEST_OPT_IN_NEGATIVE"] = "0"
        os.environ["FLYPAPER_TEST_OPT_OUT_NEGATIVE"] = "0"
        self.ev_in_not_set = os.environ.get("FLYPAPER_TEST_OPT_IN_NOT_SET")
        if self.ev_in_not_set is not None:
            del os.environ["FLYPAPER_TEST_OPT_IN_NOT_SET"]
        self.ev_out_not_set = os.environ.get("FLYPAPER_TEST_OPT_OUT_NOT_SET")
        if self.ev_out_not_set is not None:
            del os.environ["FLYPAPER_TEST_OPT_OUT_NOT_SET"]

    def tearDown(self) -> None:
        del os.environ["FLYPAPER_TEST_OPT_IN_AFFIRMATIVE"]
        del os.environ["FLYPAPER_TEST_OPT_OUT_AFFIRMATIVE"]
        del os.environ["FLYPAPER_TEST_OPT_IN_NEGATIVE"]
        del os.environ["FLYPAPER_TEST_OPT_OUT_NEGATIVE"]
        if self.ev_in_not_set is not None:
            os.environ["FLYPAPER_TEST_OPT_IN_NOT_SET"] = self.ev_in_not_set
        if self.ev_out_not_set is not None:
            os.environ["FLYPAPER_TEST_OPT_OUT_NOT_SET"] = self.ev_out_not_set

    def test_base_consent_affirmative(self):
        consent_state = consent.FlypaperConsent(True)
        self.assertTrue(consent_state.check())

    def test_base_consent_negative(self):
        consent_state = consent.FlypaperConsent(False)
        self.assertFalse(consent_state.check())

    def test_dynamic_consent_envvar_opt_in_affirmative(self):
        consent_state = consent.FlypaperConsent(
            True,
            consent.environment_variable_opt_in(
                "FLYPAPER_TEST_OPT_IN_AFFIRMATIVE", ["true", "1"]
            ),
        )
        self.assertTrue(consent_state.check())

    def test_dynamic_consent_envvar_opt_in_negative(self):
        consent_state = consent.FlypaperConsent(
            True,
            consent.environment_variable_opt_in(
                "FLYPAPER_TEST_OPT_IN_NEGATIVE", ["true", "1"]
            ),
        )
        self.assertFalse(consent_state.check())

    def test_dynamic_consent_envvar_opt_in_not_set(self):
        consent_state = consent.FlypaperConsent(
            True,
            consent.environment_variable_opt_in(
                "FLYPAPER_TEST_OPT_IN_NOT_SET", ["true", "1"]
            ),
        )
        self.assertFalse(consent_state.check())

    def test_dynamic_consent_envvar_opt_out_affirmative(self):
        consent_state = consent.FlypaperConsent(
            True,
            consent.environment_variable_opt_out(
                "FLYPAPER_TEST_OPT_OUT_AFFIRMATIVE", ["true", "1"]
            ),
        )
        self.assertFalse(consent_state.check())

    def test_dynamic_consent_envvar_opt_out_negative(self):
        consent_state = consent.FlypaperConsent(
            True,
            consent.environment_variable_opt_out(
                "FLYPAPER_TEST_OPT_OUT_NEGATIVE", ["true", "1"]
            ),
        )
        self.assertTrue(consent_state.check())

    def test_dynamic_consent_envvar_opt_out_not_set(self):
        consent_state = consent.FlypaperConsent(
            True,
            consent.environment_variable_opt_out(
                "FLYPAPER_TEST_OPT_OUT_NOT_SET", ["true", "1"]
            ),
        )
        self.assertTrue(consent_state.check())

    def test_dynamic_consent_all_allow(self):
        consent_state = consent.FlypaperConsent(
            True,
            consent.environment_variable_opt_in(
                "FLYPAPER_TEST_OPT_IN_AFFIRMATIVE", ["1"]
            ),
            consent.environment_variable_opt_out(
                "FLYPAPER_TEST_OPT_OUT_NEGATIVE", ["1"]
            ),
        )
        self.assertTrue(consent_state.check())

    def test_dynamic_consent_middle_reject_opt_in_var(self):
        consent_state = consent.FlypaperConsent(
            True,
            consent.environment_variable_opt_in("FLYPAPER_TEST_OPT_IN_NEGATIVE", ["1"]),
            consent.environment_variable_opt_out(
                "FLYPAPER_TEST_OPT_OUT_NEGATIVE", ["1"]
            ),
        )
        self.assertFalse(consent_state.check())

    def test_dynamic_consent_middle_reject_opt_out_var(self):
        consent_state = consent.FlypaperConsent(
            True,
            consent.environment_variable_opt_in(
                "FLYPAPER_TEST_OPT_IN_AFFIRMATIVE", ["1"]
            ),
            consent.environment_variable_opt_out(
                "FLYPAPER_TEST_OPT_OUT_AFFIRMATIVE", ["1"]
            ),
        )
        self.assertFalse(consent_state.check())


if __name__ == "__main__":
    unittest.main()



"""
This module implements Flypaper's user consent mechanisms.
"""
import os
from typing import Callable, cast, Sequence, Union

ConsentMechanism = Callable[[], bool]


class FlypaperConsent:
    """
    FlypaperConsent stores the client's consent settings.
    """

    def __init__(self, *mechanisms: Union[bool, ConsentMechanism]) -> None:
        if not mechanisms:
            mechanisms = (False,)
        self._mechanisms = mechanisms

    def check(self) -> bool:
        """
        Checks if all consent mechanisms signal the user's consent. If any of them signal False, returns
        False. Otherwise, returns True.
        """
        for mechanism in self._mechanisms:
            if mechanism is True:
                continue
            elif mechanism is False:
                return False
            elif not cast(ConsentMechanism, mechanism)():
                return False
        return True


def environment_variable_opt_in(
    varname: str, opt_in_values: Sequence[str]
) -> ConsentMechanism:
    def mechanism() -> bool:
        if os.environ.get(varname) in opt_in_values:
            return True
        return False

    return mechanism


def environment_variable_opt_out(
    varname: str, opt_out_values: Sequence[str]
) -> ConsentMechanism:
    def mechanism() -> bool:
        if os.environ.get(varname) in opt_out_values:
            return False
        return True

    return mechanism



"""
This module implements all Flypaper methods related to generating reports and publishing them to
Bugout knowledge bases.
"""
from dataclasses import dataclass, field
import textwrap
import traceback
from typing import List, Optional

from bugout.app import Bugout

from .consent import FlypaperConsent
from .system_information import (
    SystemInformation,
    generate as generate_system_information,
)


@dataclass
class Report:
    title: str
    content: str
    tags: List[str] = field(default_factory=list)


class Reporter:
    def __init__(
        self,
        name: str,
        consent: FlypaperConsent,
        client_id: Optional[str] = None,
        system_information: Optional[SystemInformation] = None,
        bugout_token: Optional[str] = None,
        bugout_journal_id: Optional[str] = None,
        timeout_seconds: int = 10,
    ):
        self.name = name
        self.consent = consent
        self.client_id = client_id
        if system_information is None:
            system_information = generate_system_information()
        self.system_information = system_information
        self.bugout = Bugout()
        self.bugout_token = bugout_token
        self.bugout_journal_id = bugout_journal_id
        self.timeout_seconds = timeout_seconds

    def system_tags(self) -> List[str]:
        tags = [
            "flypaper",
            "source:{}".format(self.name),
            "os:{}".format(self.system_information.os),
            "arch:{}".format(self.system_information.machine),
            "python:{}".format(self.system_information.python_version_major),
            "python:{}.{}".format(
                self.system_information.python_version_major,
                self.system_information.python_version_minor,
            ),
            "python:{}".format(self.system_information.python_version),
        ]
        if self.client_id is not None:
            tags.append("client:{}".format(self.client_id))

        return tags

    def publish(self, report: Report) -> None:
        if not self.consent.check():
            return
        if self.bugout_token is None or self.bugout_journal_id is None:
            return

        try:
            report.tags = list(set(report.tags))
            self.bugout.create_entry(
                self.bugout_token,
                self.bugout_journal_id,
                title=report.title,
                content=report.content,
                tags=report.tags,
                timeout=self.timeout_seconds,
            )
        except:
            pass

    def system_report(
        self, tags: Optional[List[str]] = None, publish: bool = True
    ) -> Report:
        title = "{}: System information".format(self.name)
        content = textwrap.dedent(
            """
            ### OS
            ```
            {os}
            ```

            Release: `{os_release}`

            ### Processor
            ```
            {machine}
            ```

            ### Python
            ```
            {python_version}
            ```
        """.format(
                os=self.system_information.os,
                os_release=self.system_information.os_release,
                machine=self.system_information.machine,
                python_version=self.system_information.python_version,
            )
        )
        report = Report(title=title, content=content, tags=self.system_tags())
        if tags is not None:
            report.tags.extend(tags)

        if publish:
            self.publish(report)

        return report

    def error_report(
        self, error: Exception, tags: Optional[List[str]] = None, publish: bool = True
    ) -> Report:
        title = "{} - {}".format(self.name, type(error).__name__)
        system_report = self.system_report(publish=False)
        error_content = textwrap.dedent(
            """
            ### Exception summary
            ```
            {error_summary}
            ```

            ### Traceback
            ```
            {error_traceback}
            ```
        """.format(
                error_summary=repr(error),
                error_traceback="".join(
                    traceback.format_exception(
                        etype=type(error),
                        value=error,
                        tb=error.__traceback__,
                    )
                ),
            )
        )
        content = "\n\n".join(
            [
                "## Error information",
                error_content,
                "- - -",
                "## System information",
                system_report.content,
            ]
        )
        if tags is None:
            tags = []
        tags.extend(system_report.tags)

        report = Report(title=title, content=content, tags=tags)

        if publish:
            self.publish(report)

        return report
